"""Price impact calculations.


"""
import enum

import pandas as pd

from dataclasses import dataclass, asdict

from tradingstrategy.liquidity import GroupedLiquidityUniverse
from tradingstrategy.types import PrimaryKey, USDollarAmount


class LiquidityDataMissing(Exception):
    """We try to get a price impact for a pair for which we have no data."""


class NoTradingPair(LiquidityDataMissing):
    """Trading pair is missing."""


class SampleTooFarOff(LiquidityDataMissing):
    """There are no samples in this timeset for a specific timepoint."""



class LiquiditySampleMeasure(enum.Enum):
    """For liquidity samples, which measurement we use for a price impact."""
    open = "open"
    close = "close"
    high = "high"
    low = "low"


@dataclass
class PriceImpact:
    """How much price impact a potential trade would have.

    Because how Uniswap v2 like DEXes operate, liquidity provider
    and protocol fees are included within the price impact calculations.
    Depending on the example these fees are not included in the calculations.

    This `PriceImpact` data also includes separate liquidity provider and protocol fees,
    as for trading companies these fees might be tax deductible.

    TODO: These fees are not yet confirmed with a live exchange.
    """

    #: Liquidity that was used for the price impact calculation, as expressed US dollars of one sided liquidity, see :py:class:`XYLiquidity`.
    available_liquidity: USDollarAmount

    #: How much % worse execution you get because due to in availability of the liquidity.
    price_impact: float

    #: How much worth of tokens you actually get for trade amount, expressed as US dollar
    delivered: USDollarAmount

    #: How much LP fees are paid in this transaction
    lp_fees_paid: USDollarAmount

    #: How much protocol fees are paid in this transaction
    protocol_fees_paid: USDollarAmount

    #: How much the trade cost you totally (trade amount - delivered)
    #: This includes LP fees paid, protocol fees paid and any loss because of limited liquidity.
    #: This does not include gas fees (tx fees) for the network.
    cost_of_trade: USDollarAmount


def estimate_xyk_price_impact(liquidity: float, trade_amount: float, lp_fee: float, protocol_fee: float) -> PriceImpact:
    """Calculates XY liquidity model slippage.

    Works for all Uniswap v2 style DEXes.

    For us, price impacts are easier because we operate solely in the US dollar space.

    Some price impact calculation examples to study:

    https://help.sushidocs.com/products/sushiswap-exchange

    https://ethereum.stackexchange.com/a/111334/620

    https://dailydefi.org/articles/price-impact-and-how-to-calculate/

    https://ethereum.stackexchange.com/a/111334/620

    TODO: Check that calculations are consistent with SushiSwap.

    :param liquidity: Liquidity expressed as USD :term:`XY liquidity model` single side liquidity.
    :param trade_amount: How much buy/sell you are doing
    :param lp_fee: Liquidity provider fee set for the pool as %. E.g. 0.0035 for Sushi.
    :param protocol_fee: Fees paid to the protocol, e.g. xSushi stakers
    """

    #price_impact = amount_in_with_fee / (liquidity + amount_in_with_fee)

    # We value both halves of the liquidity pool with
    # our precalculated US dollar reference exchange rate,
    # and thus skip a lot of uint256 bit issues what comes when dealing with raw
    # tokens amounts.
    total_fees = protocol_fee + lp_fee
    reserve_a_initial = reserve_b_initial = liquidity
    amount_in_with_fee = trade_amount * (1 - total_fees)

    lp_fees_paid = trade_amount * lp_fee
    protocol_fees_paid = trade_amount * protocol_fee

    constant_product = reserve_a_initial * reserve_b_initial
    reserve_b_after_execution = constant_product / (reserve_a_initial + amount_in_with_fee)
    amount_out = reserve_b_initial - reserve_b_after_execution
    market_price = amount_in_with_fee / amount_out
    mid_price = reserve_a_initial / reserve_b_initial
    price_impact = 1 - (mid_price / market_price)
    price_impact = abs(price_impact)
    cost_of_trade = trade_amount - amount_out

    return PriceImpact(
        available_liquidity=liquidity,
        delivered=amount_out,
        price_impact=price_impact,
        lp_fees_paid=lp_fees_paid,
        protocol_fees_paid=protocol_fees_paid,
        cost_of_trade=cost_of_trade,
    )


class HistoricalXYPriceImpactCalculator:
    """Calculates the Uniswap slippage, old price and new price based on :term:`XY liquidity model`.

    Used for backtesting and historical price impact calculations.

    The price impact model here is naive. It assumes that the trade would only use assets in a single pool.
    However, for any real DEX this is not the case. All DEXes implement the equivalent of
    `Uniswap auto router <https://uniswap.org/blog/auto-router>`_ also known as smart order routing (SOR).
    Routing finds the optimal path for the swaps between different tokens and can include three hop
    trades or even four hop trades to find the best price for the swapper. Thus, in real life
    the price impact might be less than what this model gives to you.

    TODO: These fees are not yet confirmed with a live exchange.

    .. note ::

        Currently we do not have dynamic liquidity provider `lp_fees` data for all the pairs, as it may vary
        pair by pair. Thus, in your model you need to manually confirm you are using the correct `lp_fees` value.
    """

    def __init__(self, liquidity_universe: GroupedLiquidityUniverse, lp_fee=0.0030, protocol_fee=0):
        """
        :param lp_fees: Liquidity provider fees as 0...1 % number
        """
        self.liquidity_universe = liquidity_universe
        # TODO: Later, pull this data dynamicalyl from the exchanges
        self.lp_fee = lp_fee
        self.protocol_fee = protocol_fee

    def calculate_price_impact(self, when: pd.Timestamp, pair_id: PrimaryKey, trade_amount: USDollarAmount, measurement: LiquiditySampleMeasure=LiquiditySampleMeasure.open, max_distance: pd.Timedelta=pd.Timedelta(days=1)) -> PriceImpact:
        """What would have been a price impact if a Uniswap-style trade were executed in the past.

        :param measurement: By default, we check the liquidity based on the liquidity available at the sample openining time.

        :param max_distance: If the sample is too far off, then abort because of gaps in data. Depending on the candle time frame you operate, you might need to adjust `max_distance`  to account the sample timestamp skew. For example, if you are using weekly candles, you need to set this to seven days.

        """

        liquidity_samples = self.liquidity_universe.get_liquidity_samples_by_pair(pair_id)
        if liquidity_samples is None:
            raise NoTradingPair(f"The universe does not contain liquidity data for pair {pair_id}")

        # measurement_samples = liquidity_samples[when:]

        ranged_samples = liquidity_samples[when:]

        if len(ranged_samples) == 0:
            raise SampleTooFarOff(f"Pair {pair_id} has no liquidity samples before {when}")

        # pair_id                        74846
        # timestamp        2021-06-07 00:00:00
        # exchange_rate            2489.434326
        # open                     259676608.0
        # close                    191939088.0
        # high                     267280096.0
        # low                      172410688.0
        # adds                             249
        # removes                          154
        # syncs                           7824
        # add_volume                2689543.75
        # remove_volume             30361768.0
        # start_block                 12584093
        # end_block                   12629257
        # Name: 2021-06-07 00:00:00, dtype: object

        first_sample = ranged_samples.iloc[0]

        ts = first_sample.timestamp

        distance = abs(ts - when)
        if distance > max_distance:
            raise SampleTooFarOff(f"Pair {pair_id} has liquidity samples, but the sample we got at {ts} is too far off from {when}. Distance is {distance} when we want at least {max_distance}")

        # "open", "close", etc.
        liquidity_at_sample = first_sample[measurement.value]

        return estimate_xyk_price_impact(liquidity_at_sample, trade_amount, self.lp_fee, self.protocol_fee)








"""Cached data store for trade feeds.

We store

- Block headers

- Fetched trades

We do not store

- Candles (always regenerated)
"""
import logging
import os
import shutil
from pathlib import Path
from typing import Tuple

from eth_defi.event_reader.block_header import BlockHeader
from eth_defi.event_reader.parquet_block_data_store import ParquetDatasetBlockDataStore
from tradingstrategy.direct_feed.trade_feed import TradeFeed


logger = logging.getLogger(__name__)


class DirectFeedStore:
    """Manage on-disk block header and trade cache for direct feeds.

    Internally uses partitioned Parquet dataset storage.
    Each partition is a range of blocks and goes to different folder/file.
    """

    def __init__(self, base_path: Path, partition_size: int):
        """Initialise a new store.

        :param base_path:
            Base folder where data is dumped.
            Both headers and trades get their own Parquet datasets
            as folders.

        :param partition_size:
            Partition size for the store.
            Expressed as number of blocks per parquet file.

        """
        assert isinstance(base_path, Path)
        assert type(partition_size) == int
        self.base_path = base_path
        self.partition_size = partition_size

    def is_empty(self) -> bool:
        """Have we written anything to this store yer."""
        return not self.base_path.exists()

    def clear(self):
        """Clear cache."""
        assert not self.is_empty(), f"Cannot clear empty store."
        if self.base_path.exists():
            shutil.rmtree(self.base_path)

    def save_trade_feed(self, trade_feed: TradeFeed) -> Tuple[int, int]:
        """Save the trade and block header data.

        :param trade_feed:
            Save trades and block headers from this feed.

        :return:
            Last saved header block number, last saved trade number
        """

        base_path = self.base_path
        partition_size = self.partition_size

        header_store = ParquetDatasetBlockDataStore(Path(base_path).joinpath("blocks"), partition_size)
        trade_store = ParquetDatasetBlockDataStore(Path(base_path).joinpath("trades"), partition_size)

        # Save headers
        headers_df = trade_feed.reorg_mon.to_pandas(partition_size)
        if len(headers_df) > 0:
            header_store.save(headers_df)
            assert not header_store.is_virgin(), f"Headers not correctly written"
            last_header_block = headers_df.iloc[-1]["block_number"]
        else:
            last_header_block = 0

        # Save trades
        trades_df = trade_feed.to_pandas(partition_size)
        if len(trades_df) > 0:
            trade_store.save(trades_df, check_contains_all_blocks=False)
            assert not trade_store.is_virgin(), f"Trades not correctly written"
            last_trade_block = trades_df.iloc[-1]["block_number"]
        else:
            last_trade_block = 0

        return last_header_block, last_trade_block

    def load_trade_feed(self, trade_feed: TradeFeed) -> bool:
        """Load trade and block header data.

        :param trade_feed:
            Save trades and block headers from this feed.

        :return:
            True if any data was loaded.
        """

        base_path = self.base_path
        partition_size = self.partition_size

        header_store = ParquetDatasetBlockDataStore(Path(base_path).joinpath("blocks"), partition_size)
        trade_store = ParquetDatasetBlockDataStore(Path(base_path).joinpath("trades"), partition_size)

        if not header_store.is_virgin():
            logger.info("Loading block header data from %s", header_store.path)
            headers_df_2 = header_store.load()
            block_map = BlockHeader.from_pandas(headers_df_2)
            trade_feed.reorg_mon.restore(block_map)
            logger.info("Loaded %d blocks", len(block_map))

        if not trade_store.is_virgin():
            trades_df_2 = trade_store.load()
            trade_feed.restore(trades_df_2)
            logger.info(f"Loaded {len(trades_df_2)}, last block is {trade_feed.get_block_number_of_last_trade():,}")

        return not(header_store.is_virgin() or trade_store.is_virgin())



"""Synthetic random price data feed for unit testing."""

import random
from decimal import Decimal
from typing import Dict, Optional, List, Iterable, Type

import pandas as pd
from tqdm import tqdm

from eth_defi.event_reader.block_header import BlockHeader
from eth_defi.price_oracle.oracle import BasePriceOracle
from eth_defi.event_reader.reorganisation_monitor import ReorganisationMonitor

from .timeframe import Timeframe
from .trade_feed import TradeFeed, Trade
from .direct_feed_pair import PairId


class SyntheticTradeFeed(TradeFeed):
    """Synthetic trade feed that produces random trades."""

    def __init__(self,
                 pairs: List[PairId],
                 oracles: Dict[PairId, BasePriceOracle],
                 reorg_mon: ReorganisationMonitor,
                 data_retention_time: Optional[pd.Timedelta] = None,
                 random_seed = 1,
                 start_price_range=150,
                 end_price_range=300,
                 min_trades_per_block=0,
                 max_trades_per_block=10,
                 min_amount=50,
                 max_amount=50,
                 price_movement_per_trade=2.5,
                 timeframe: Timeframe = Timeframe("1min"),
                 prices: Dict[PairId, float] = None,
                 ):
        super().__init__(
            pairs=pairs,
            oracles=oracles,
            reorg_mon=reorg_mon,
            data_retention_time=data_retention_time,
            timeframe=timeframe,
        )

        self.pairs = pairs
        self.min_trades_per_block = min_trades_per_block
        self.max_trades_per_block = max_trades_per_block
        self.min_amount = min_amount
        self.max_amount = max_amount
        self.price_movement_per_trade = price_movement_per_trade
        self.random_gen = random.Random(random_seed)

        if not prices:
            self.prices = {
                p: self.random_gen.randint(start_price_range, end_price_range)
                for p in pairs
            }
        else:
            self.prices = prices

    def fetch_trades(self, start_block: int, end_block: Optional[int], tqdm: Optional[Type[tqdm]] = None) -> Iterable[Trade]:
        """Generate few random trades per block per pair."""

        block_data = {b.block_number: b for b in self.reorg_mon.fetch_block_data(start_block, end_block)}

        max_blocks = end_block - start_block

        if max_blocks > 5 and tqdm:
            progress_bar = tqdm(total=max_blocks)
        else:
            progress_bar = None

        for block_num in range(start_block, end_block + 1):
            for p in self.pairs:
                trades_per_block = self.random_gen.randint(self.min_trades_per_block, self.max_trades_per_block)
                for trade_idx in range(trades_per_block):

                    self.prices[p] += self.random_gen.uniform(-self.price_movement_per_trade, self.price_movement_per_trade)
                    self.prices[p] = max(self.prices[p], 0.00001)  # Don't go to negative prices
                    price = self.prices[p]
                    amount = self.random_gen.uniform(self.min_amount, self.max_amount)

                    block: BlockHeader = block_data[block_num]
                    tx_hash = hex(self.random_gen.randint(2**31, 2**32))
                    log_index = trade_idx

                    exchange_rate = self.get_exchange_rate(p)

                    yield Trade(
                           pair=p,
                           block_number=block_num,
                           block_hash=block.block_hash,
                           timestamp=pd.Timestamp.utcfromtimestamp(block.timestamp),
                           tx_hash=tx_hash,
                           log_index=log_index,
                           price=Decimal(price),
                           amount=Decimal(amount),
                           exchange_rate=exchange_rate,
                        )

            if progress_bar:
                progress_bar.update(1)

        if progress_bar:
            progress_bar.close()



"""Time window presentation."""
import datetime
import enum

import pandas as pd
from pandas.tseries.frequencies import to_offset


class TimeBucket(enum.Enum):
    """Supported time windows for :term:`candle` and :term:`liquidity` data.

    We use term "bucket", from the
    `TimescaleDB slang <https://docs.timescale.com/api/latest/continuous-aggregates/refresh_continuous_aggregate/>`_
    to symbol the time window of a candle data we are querying.

    The raw blockchain data is assembled to 1 minute time buckets.
    Then the 1 minute timebuckets are resampled to other windows.

    All time windows are in UTC.
    Daily time buckets have their hour, minute and second set to the zero in the outputted data.
    Hourly time buckets have minute and hour set to zero, etc.

    Python labels are reserved from the actual values, because Python symbol cannot start with a number.
    """

    #: One minute candles
    m1 = "1m"

    #: Five minute candles
    m5 = "5m"

    #: Quarter candles
    m15 = "15m"

    #: Hourly candles
    h1 = "1h"

    #: Four hour candles
    h4 = "4h"

    #: Daily candles
    d1 = "1d"

    #: Weekly candles
    d7 = "7d"

    #: Monthly candles
    d30 = "30d"

    #: We do not have "yearly" candles, but some trade statistics are calculated
    #: for 360 days, thus we need a corresponding time bucket for them.
    d360 = "360d"

    #: Some statistics like "all time high", for example, only make sense if a "bucket"
    #: spans across the entire timeline.
    infinite = "infinite"

    #: A placeholder value representing a "NULL value" for cases where Python's None
    #: is not a favorable choice for some reason.
    not_applicable = "not_applicable"

    def to_timedelta(self) -> datetime.timedelta:
        """Get delta object for a TimeBucket definition.

        You can use this to construct arbitrary timespans or iterate candle data.
        """
        return _DELTAS[self]

    def to_pandas_timedelta(self) -> pd.Timedelta:
        """Get pandas delta object for a TimeBucket definition.

        You can use this to construct arbitrary timespans or iterate candle data.
        """
        return pd.Timedelta(_DELTAS[self])

    def to_frequency(self) -> pd.DateOffset:
        """Get frequency input for Pandas fuctions.

        You can use this to construct arbitrary timespans or iterate candle data.
        """
        if self in {TimeBucket.infinite, TimeBucket.not_applicable}:
            raise ValueError(f"Enum member {self} cannot be mapped to a frequency.")

        delta = self.to_timedelta()
        return to_offset(delta)


# datetime.timedelta equivalents of different time buckets
_DELTAS = {
    TimeBucket.m1: datetime.timedelta(minutes=1),
    TimeBucket.m5: datetime.timedelta(minutes=5),
    TimeBucket.m15: datetime.timedelta(minutes=15),
    TimeBucket.h1: datetime.timedelta(hours=1),
    TimeBucket.h4: datetime.timedelta(hours=4),
    TimeBucket.d1: datetime.timedelta(days=1),
    TimeBucket.d7: datetime.timedelta(days=7),
    TimeBucket.d30: datetime.timedelta(days=30),
    TimeBucket.d360: datetime.timedelta(days=360),
    TimeBucket.infinite: datetime.timedelta.max,
    TimeBucket.not_applicable: datetime.timedelta(0),  # some sort of a NULL value
}





"""Load candle and liquidity data over Trading Strategy real-time API.

[Use JSONL transport](https://tradingstrategy.ai/api/explorer/).
This method does not require API key at the moment.
"""
import logging
from collections import defaultdict

import pandas as pd

import datetime
from typing import List, Optional, Dict, Set

import requests
import jsonlines
from numpy import NaN
from tqdm_loggable.auto import tqdm

from tradingstrategy.candle import Candle
from tradingstrategy.timebucket import TimeBucket
from tradingstrategy.utils.time import to_int_unix_timestamp

logger = logging.getLogger(__name__)


#: Column name mappings from JSONL data to our :py:class:`~tradingstrategy.candle.Candle`.
CANDLE_MAPPINGS = {
    "ci": None,  # chain_id discarded
    "ei": None,  # exchange_id discarded
    "p": "pair_id",
    "ts": "timestamp",
    "o": "open",
    "h": "high",
    "l": "low",
    "c": "close",
    "xr": "exchange_rate",
    "b": "buys",
    "s": "sells",
    "bv": "buy_volume",
    "sv": "sell_volume",
    "sb": "start_block",
    "eb": "end_block",
    "tc": None,  # Currently not available for Uni v3 exchanges
    "v": None,  # Deprecated
}


class JSONLMaxResponseSizeExceeded(Exception):
    """Raised if we ask too much JSONL data from the server."""


class NoJSONLData(Exception):
    """Server did not return any data for some reason."""


def load_trading_strategy_like_jsonl_data(
    session: requests.Session,
    api_url: str,
    pair_ids: Set[int],
    time_bucket: TimeBucket,
    mappings: Dict[str, str],
    start_time: Optional[datetime.datetime] = None,
    end_time: Optional[datetime.datetime] = None,
    max_bytes: Optional[int] = None,
    progress_bar_description: Optional[str] = None,
) -> pd.DataFrame:
    """Read data from JSONL endpoint.

    `See OpenAPI spec for details on the format <https://tradingstrategy.ai/api/explorer/>`_.

    Can be used to load

    - OHLCV candles

    - Liquidity data

    Calling this function may consume up to few hundred megabytes
    of memory depending on the response size.

    Display a progress bar using :py:mod:`tqdm`.

    :param df:
        The master DataFrame we are going to fill up.

    :param api_url:
        Which Trading Strategy API we call

    :param pair_ids:
        Trading pairs we query data for

    :param time_bucket:
        Candle time frame

    :param mappings:
        Mapping between JSONL object keys and DataFrame columns

    :return:
        In-place modified DataFrame passed to this function
    """

    params = {
        "pair_ids": ",".join(str(i) for i in pair_ids),
        "time_bucket": time_bucket.value,
    }

    if start_time:
        params["start"] = start_time.isoformat()

    if end_time:
        params["end"] = end_time.isoformat()

    if max_bytes:
        params["max_bytes"] = max_bytes

    logger.info("Loading JSON data, endpoint:%s, params:%s", api_url, params)

    resp = session.get(api_url, params=params, stream=True)
    reader = jsonlines.Reader(resp.raw)

    candle_data = defaultdict(list)

    # Figure out how to plot candle download progress using TQDM
    # Draw progress bar using timestamps first candle - last candle
    progress_bar_start = None
    progress_bar_end = end_time or datetime.datetime.utcnow()
    progress_bar_end = to_int_unix_timestamp(progress_bar_end)
    current_ts = last_ts = None
    progress_bar = None
    refresh_rate = 200  # Update the progress bar for every N candles

    # Massage the format good for pandas
    for idx, item in enumerate(reader):

        # Stream terminated
        if "error" in item:
            raise JSONLMaxResponseSizeExceeded(str(item))

        current_ts = item["ts"]

        # Set progress bar start to the first timestamp
        if not progress_bar_start and progress_bar_description:
            progress_bar_start = current_ts
            logger.debug("First candle timestamp at %s", current_ts)
            total = progress_bar_end - progress_bar_start
            assert progress_bar_start <= progress_bar_end, f"Mad progress bar {progress_bar_start} - {progress_bar_end}"
            progress_bar = tqdm(desc=progress_bar_description, total=total)

        # Translate the raw compressed keys to our internal
        # Pandas keys
        for key, value in item.items():
            translated_key = mappings[key]
            if translated_key is None:
                # Deprecated/discarded keys
                continue

            candle_data[translated_key].append(value)

        if idx % refresh_rate == 0:
            if last_ts and progress_bar:
                progress_bar.update(current_ts - last_ts)
                progress_bar.set_postfix({"Currently at": datetime.datetime.utcfromtimestamp(current_ts)})
            last_ts = current_ts

    # Some data validation facilities
    assumed_lenght = None
    previous_key = None
    for key, arr in candle_data.items():
        current_array_len = len(arr)
        if assumed_lenght is None:
            assumed_lenght = current_array_len
        elif assumed_lenght != current_array_len:
            raise RuntimeError(f"Bad JSONL data. The length for {previous_key} was {assumed_lenght}, but the length for {key} is {current_array_len}")
        previous_key = key

    if progress_bar:
        # https://stackoverflow.com/a/45808255/315168
        if current_ts and last_ts:
            progress_bar.update(current_ts - last_ts)
        progress_bar.close()

    if len(candle_data) == 0:
        raise NoJSONLData(f"Did not get any data, url:{api_url}, params:{params}")

    df = pd.DataFrame.from_dict(candle_data)

    logger.debug("Loaded %d rows", len(df))
    return df


def load_candles_jsonl(
    session: requests.Session,
    server_url: str,
    pair_ids: Set[id],
    time_bucket: TimeBucket,
    start_time: Optional[datetime.datetime] = None,
    end_time: Optional[datetime.datetime] = None,
    max_bytes: Optional[int] = None,
    progress_bar_description: Optional[str] = None,
    sanity_check_count=75,
) -> pd.DataFrame:
    """Load candles using JSON API and produce a DataFrame.

    Serially load each pair data.

    - Load data from per-pair JSON endpoint
    - Each pair becomes pandas :py:class:`pd.Series`
    - The final DataFrame is the merge of these series

    See :py:mod:`tradingstrategy.candle` for candle format description.

    :param progress_bar_description:
        Progress bar label

    :param sanity_check_count:
        Don't accidentally try to load too many pairs.

        Never exceed this value.

    :raise JSONLMaxResponseSizeExceeded:
        If the max_bytes limit is breached

    :return:
        Dataframe with candle data for giving pairs.
    """

    assert len(pair_ids) < sanity_check_count, f"We have {len(pair_ids)}, but you probably don't want to use this data fetch method for more than {sanity_check_count} pairs"

    api_url = f"{server_url}/candles-jsonl"

    df = load_trading_strategy_like_jsonl_data(
        session,
        api_url,
        pair_ids,
        time_bucket,
        CANDLE_MAPPINGS,
        start_time,
        end_time,
        max_bytes,
        progress_bar_description,
    )

    # Not supported at the momemnt
    df.loc[:, "avg"] = NaN

    df = df.astype(Candle.DATAFRAME_FIELDS)

    # Reconstruct normal volume column
    # as expected for OHLCV data
    df["volume"] = df["buy_volume"] + df["sell_volume"]

    # Convert JSONL unix timestamps to Pandas
    df["timestamp"] = pd.to_datetime(df['timestamp'], unit='s')

    # Assume candles are always indexed by their timestamp
    df.set_index("timestamp", inplace=True, drop=False)

    return df



"""Schema manipulation utitilies.

For more information about Pyarrow Schemas, see https://arrow.apache.org/docs/python/api/datatypes.html
"""

import typing
from enum import Enum
from typing import Optional, Dict, Callable, List
from dataclasses import fields, Field

import pyarrow as pa

from tradingstrategy.chain import ChainId
from tradingstrategy.types import PrimaryKey, NonChecksummedAddress, BlockNumber, UNIXTimestamp, BasisPoint


class CannotMap(Exception):
    pass


def unmappable(t):
    raise CannotMap(f"Cannot automatically map {t}")


#: Default mappings for automatic schema generation,
#: Including our own type definitions,
DEFAULT_MAPPINGS = {
    PrimaryKey: lambda t: pa.uint32(),
    ChainId: lambda t: pa.uint16(),
    NonChecksummedAddress: lambda t: pa.string(),
    BlockNumber: lambda t: pa.uint32(),
    UNIXTimestamp: lambda t: pa.timestamp("s"),
    BasisPoint: lambda t: pa.uint32(),
    bool: lambda t: pa.bool_(),
    float: lambda t: pa.float32(),
    int: lambda t: pa.uint32(),
    str: lambda t: pa.string(),
    dict: lambda t: unmappable(t),
    list: lambda t: unmappable(t),
}


def map_field_to_arrpw(field: Field, hints: Dict[str, pa.DataType], core_mappings: Dict[str, Callable]) -> pa.DataType:
    """Map a dataclass field to a pyarrow equivalent, respect hints"""
    hinted = hints.get(field.name)
    if hinted:
        return hinted

    # Resolve optional
    origin = typing.get_origin(field.type)
    if origin == typing.Union:
        # Optional type
        args = typing.get_args(field.type)
        assert len(args) == 2
        assert args[1] == type(None)
        field_type = args[0]
        true_origin = typing.get_origin(args[0])
    else:
        field_type = field.type
        true_origin = origin

    if true_origin == list:
        args = typing.get_args(field_type)
        value_type = args[0]
        mapped_value_type = core_mappings[value_type](field)
        return pa.list_(mapped_value_type)
    elif true_origin == dict:
        # Only string string dicts supported
        return pa.map_(pa.string(), pa.string())
    else:

        if issubclass(field_type, Enum):
            # No support for category compaction yet
            field_type = str

        return core_mappings[field_type](field)


def create_pyarrow_schema_for_dataclass(
        cls,
        hints: Optional[typing.Dict[str, pa.DataType]] = None,
        core_mappings=DEFAULT_MAPPINGS) -> pa.Schema:
    """Map a Python dataclass to Pyarrow schema.

    Most fields map automatically, but you can also provide per field name hints what types they should use.
    """

    if not hints:
        hints = {}

    pa_fields = [(field.name, map_field_to_arrpw(field, hints, core_mappings)) for field in fields(cls)]
    return pa.schema(pa_fields)


def create_columnar_work_buffer(cls) -> Dict[str, list]:
    """Create a columnar work buffer to export data into Pyarrow Tables."""
    return {field.name: [] for field in fields(cls)}


def append_to_columnar_work_buffer(buffer: Dict[str, list], item):
    """Convert tabular data items to columnar.

    Automatically handle the special case of enum.
    """

    def process_value(key: str):
        try:
            v = getattr(item, key)
            if isinstance(v, Enum):
                v = v.value
            else:
                v = v
            buffer[key].append(v)
        except (AttributeError, ValueError) as e:
            raise RuntimeError(f"Could not serialised {key} for {item}") from e

    for key in buffer:
        process_value(key)


"""Generic units used in data models used in Capitalgram.

Types aliases are used to give human-readable meaning for various arguments and return values.
These are also used to hint :term:`Pyarrow` schemas to make :term:`Parquet` files more compact.
"""

from typing import TypeAlias

#: 64-bit integer based primary key.
#:
#:     Also referred as "internal id" on website and data streams.
#:
#:     This is a type alias to Python integer type.
#:
#:     Primary keys are not stable across different dataset. Blockchain data healing process
#:     may require to regenerate the data which means the old data is purged,
#:     reimported, with new primary keys.
PrimaryKey: TypeAlias = int



#: Ethereum address that does *not* use EIP-55 checksumming.
#:
#:     EIP-55 https://github.com/ethereum/EIPs/blob/master/EIPS/eip-55.md
#
NonChecksummedAddress: TypeAlias = str


#: Express USD monetary amount.
#:
#: Used in exchange rates, volumes and prices in candle data.
#; Normally you should not use float for pricing,
#: but because we are already losing a lot of accuraty in float conversion
#: and our price data is not for exact transactions, but for modelling,
#: this is ok.
#:
USDollarAmount: TypeAlias = float

#: Seconds since 1.1.1970 as UTC time as integer
UNIXTimestamp: TypeAlias = int


#: EVM block number from 1 to infinity
BlockNumber: TypeAlias = int

#: Multiplier as 1/10000 or 0.01%
BasisPoint: TypeAlias = int

#: Chain id that is not a wrapped enum.
#:
#: See :py:class:`tradingstategy.chain.ChainId` for details
RawChainId: TypeAlias = int


#: Percent expressed as float 0...1
#:
#: 1 = 100%
#:
#: 0.00030 = 0.3% = 30 BPS
#:
#: See also
#:
#: - :py:data:`BasisPoint`
#:
Percent: TypeAlias = float
